#include "include/common.h"

static char	*admin_sql[] = {
	"KILL ",
	"CREATE DATABASE ",
	"DROP DATABASE ",
	"ALTER USER ",
	"CREATE USER ",
	"DROP USER ",
	"GRANT ",
	"REVOKE "
};

static int	admin_sql_len[] = {
	5, 16, 14, 11, 12, 10, 6, 7
};

static int	admin_sql_cnt = 8;

int admin_process( struct thread_t *self, int db_open_res )
{
	char			*query;
	int			i;

	extern regex_t		reg_admin_kill;
	extern regex_t		reg_admin_createdb;
	extern regex_t		reg_admin_dropdb;
	extern regex_t		reg_admin_createuser;
	extern regex_t		reg_admin_dropuser;
	extern regex_t		reg_admin_alteruser;

	extern regex_t		reg_admin_grant;
	extern regex_t		reg_admin_revoke;
	extern regex_t		reg_admin_grant_create;
	extern regex_t		reg_admin_revoke_create;

	regmatch_t		match[10];
	int			match_res = -1;

	char			*value;
	struct db_user_t	user_data;
	struct grant_t		grant_data;
	int			result = -1;


	if ( self->request.data_len <= 0)
		return SQLITE_ERROR_ADMIN;

	query = self->request.data;
	query += strspn( query, " \r\n\t" );

	if ( query[0] == 's' || query[0] == 'S' )
		return SQLITE_ERROR_ADMIN;

#ifdef __DEBUG
	printf("***** Admin extended SQL: %s\n", query);
#endif

	for ( i=admin_sql_cnt; i--; )
	{
		if ( strncasecmp( admin_sql[i], query, admin_sql_len[i] ) == 0 )
		{
			switch (i)
			{
				case 0:
					//KILL SESSION_NO
					match_res = regexec( &reg_admin_kill, query, 2, match, 0 );
					if ( match_res == 0 )
					{
					}
					//send error and return
					return response_error( self, RESPONSE_ERR_INTERNAL, "\"KILL\" SQL is not support yet." );
				break;
				case 1:
					//CREATE DATABASE DBNAME
					match_res = regexec( &reg_admin_createdb, query, 2, match, 0 );
					if ( match_res == 0 )
					{
						value = &(query[match[1].rm_so]);
						query[match[1].rm_eo] = '\0';
						if ( db_open_init( &self->db, value, NULL ) == SQLITE_OK )
						{
							memcpy( self->db.db_name, value, self->request.header.db_name_len );
							//send ok and return
							return response_info( self, RESPONSE_INFO_DB_CREATE, NULL );
						}
					}
					//send error and return
					return response_error( self, RESPONSE_ERR_DB_CREATE, NULL );
				break;
				case 2:
					//DROP DATABASE DBNAME
					match_res = regexec( &reg_admin_dropdb, query, 2, match, 0 );
					if ( match_res == 0 )
					{
						value = &(query[match[1].rm_so]);
						query[match[1].rm_eo] = '\0';
						if ( unlink( value ) == 0 )
						{
							//send ok and return
							return response_info( self, RESPONSE_INFO_DB_DROP, NULL );
						}
					}
					//send error and return
					return response_error( self, RESPONSE_ERR_DB_DROP, NULL );
				break;
				case 3:
					//ALTER USER
					match_res = regexec( &reg_admin_alteruser, query, 10, match, 0 );
					if ( match_res == 0 )
					{
						value = &(query[match[1].rm_so]);
						query[match[1].rm_eo] = '\0';
						strncpy( user_data.id, value, 128 );

						if ( match[4].rm_so > 0 && strncasecmp( &(query[match[4].rm_so]), "PASSWORD", 8 ) == 0 )
						{
							//password change
							value = &(query[match[5].rm_so]);
							query[match[5].rm_eo] = '\0';
							if ( value[0] == '\'' && query[match[5].rm_eo-1] == '\'' )
							{
								query[match[5].rm_eo-1] = '\0';
								value++;
							}
							strncpy( user_data.pass, value, 128 );

							if ( admin_alter_user( &user_data, ADMIN_USER_MANAGE_ALTER1 ) == SQLITE_OK )
							{
								auth_user_flush();
								return response_info( self, RESPONSE_INFO_USER_ALTER, NULL );
							}
						}
						else
						{
							if ( match[7].rm_so > 0 && strncasecmp( &(query[match[7].rm_so]), "NO", 2 ) == 0 )
								user_data.super = 0;
							else
								user_data.super = 1;

							if ( admin_alter_user( &user_data, ADMIN_USER_MANAGE_ALTER2 ) == SQLITE_OK )
							{
								auth_user_flush();
								return response_info( self, RESPONSE_INFO_USER_ALTER, NULL );
							}
						}
					}
					//send error and return
					return response_error( self, RESPONSE_ERR_INTERNAL, "\"ALTER USER\" SQL syntax not valid." );
				break;
				case 4:
					//CREATE USER ID PASSWORD [SUPERUSER]
					match_res = regexec( &reg_admin_createuser, query, 5, match, 0 );
					if ( match_res == 0 )
					{
						value = &(query[match[1].rm_so]);
						query[match[1].rm_eo] = '\0';
						strncpy( user_data.id, value, 128 );

						value = &(query[match[2].rm_so]);
						query[match[2].rm_eo] = '\0';
						if ( value[0] == '\'' && query[match[2].rm_eo-1] == '\'' )
						{
							query[match[2].rm_eo-1] = '\0';
							value++;
						}
						strncpy( user_data.pass, value, 128 );

						value = &(query[match[3].rm_so]);
						query[match[3].rm_eo] = '\0';
						user_data.super = (strncasecmp(value, "SUPERUSER", 9) == 0)?1:0;

						if ( admin_create_user( &user_data ) == SQLITE_OK )
						{
							//send ok and return;
							auth_user_flush();
							return response_info( self, RESPONSE_INFO_USER_CREATE, NULL );
						}
					}
					//send error and return
					return response_error( self, RESPONSE_ERR_USER_CREATE, NULL );
				break;
				case 5:
					//DROP USER ID [DBNAME]
					match_res = regexec( &reg_admin_dropuser, query, 3, match, 0 );
					if ( match_res == 0 )
					{
						memset( &user_data, 0, sizeof(user_data) );

						value = &(query[match[1].rm_so]);
						query[match[1].rm_eo] = '\0';
						strncpy( user_data.id, value, 128 );

						if ( admin_drop_user( &user_data ) == SQLITE_OK )
						{
							//send ok and return;
							auth_user_flush();
							return response_info( self, RESPONSE_INFO_USER_DROP, NULL );
						}
					}
					//send error and return
					return response_error( self, RESPONSE_ERR_USER_DROP, NULL );
				break;
				case 6:
				case 7:
					//GRANT, REVOKE CREATE
					if ( i == 6 )
						match_res = regexec( &reg_admin_grant_create, query, 10, match, 0 );
					else
						match_res = regexec( &reg_admin_revoke_create, query, 10, match, 0 );
					if ( match_res == 0 )
					{
						//rule
						grant_data.rule = &(query[match[1].rm_so]);
						query[match[1].rm_eo] = '\0';

						grant_data.dbname = &(query[match[2].rm_so]);
						query[match[2].rm_eo] = '\0';

						grant_data.user = &(query[match[4].rm_so]);
						query[match[4].rm_eo] = '\0';

						if ( i == 6 )
							result = auth_create_update( SQLITE_AUTH_GRANT_CREATE, &grant_data );
						else
							result = auth_create_update( SQLITE_AUTH_REVOKE_CREATE, &grant_data );
					}

					//GRANT, REVOKE
					if ( i == 6 )
						match_res = regexec( &reg_admin_grant, query, 10, match, 0 );
					else
						match_res = regexec( &reg_admin_revoke, query, 10, match, 0 );
					if ( match_res == 0 )
					{
						//rule
						grant_data.rule = &(query[match[1].rm_so]);
						query[match[1].rm_eo] = '\0';

						grant_data.dbname = &(query[match[2].rm_so]);
						query[match[2].rm_eo] = '\0';

						if ( match[3].rm_so > 0 )
						{
							grant_data.object = &(query[match[3].rm_so]);
							query[match[3].rm_eo] = '\0';
						}
						else
							grant_data.object = "*";

						grant_data.user = &(query[match[5].rm_so]);
						query[match[5].rm_eo] = '\0';

						if ( i == 6 )
							result = auth_update( SQLITE_AUTH_GRANT, &grant_data );
						else
							result = auth_update( SQLITE_AUTH_REVOKE, &grant_data );
					}

					if ( result == SQLITE_OK )
					{
						if ( i == 6 )
							return response_info( self, RESPONSE_INFO_GRANT, NULL );
						else
							return response_info( self, RESPONSE_INFO_REVOKE, NULL );
					}
					//send error and return
					return response_error( self, RESPONSE_ERR_GRANT, NULL );
				break;
			}
		}
	}

	return SQLITE_ERROR_ADMIN;
}

int admin_create_user( struct db_user_t *user_data )
{
	return admin_user_manage( user_data, ADMIN_USER_MANAGE_CREATE );
}

int admin_drop_user( struct db_user_t *user_data )
{
	return admin_user_manage( user_data, ADMIN_USER_MANAGE_DROP1 );
}

int admin_alter_user( struct db_user_t *user_data, int method )
{
	return admin_user_manage( user_data, method );
}

int admin_user_manage( struct db_user_t *user_data, int method )
{
	extern struct db_t	db_auth;
	extern pthread_mutex_t	db_auth_mutex;

	char			*sql = NULL;
	int			result;

	switch ( method )
	{
		case ADMIN_USER_MANAGE_CREATE:
			sql = db_printf( ADMIN_SQL_CREATE_USER, user_data->id, user_data->pass, user_data->super );
		break;
		case ADMIN_USER_MANAGE_DROP1:
			sql = db_printf( ADMIN_SQL_DROP_USER1, user_data->id );
		break;
		case ADMIN_USER_MANAGE_ALTER1:
			sql = db_printf( ADMIN_SQL_ALTER_USER1, user_data->pass, user_data->id );
		break;
		case ADMIN_USER_MANAGE_ALTER2:
			sql = db_printf( ADMIN_SQL_ALTER_USER2, user_data->super, user_data->id );
		break;
	}
printf("%s\n", sql);
	if ( sql == NULL )
		return SQLITE_NOMEM;

	pthread_mutex_lock( &db_auth_mutex );
	result = db_throw( &db_auth, sql );
	pthread_mutex_unlock( &db_auth_mutex );
	db_free( sql );

	return result;
}

