#include "include/common.h"

void init_config()
{
	extern struct config_t		config;

	config.filepath =		ENV_DEF_FILEPATH;

	config.server_port =		ENV_DEF_PORT;
	config.max_session =		ENV_DEF_MAX_SESSION;		//

	config.poll_timer =		ENV_DEF_POLL_TIMEOUT * 1000;
	config.access_control =		ENV_DEF_ACCESS_CONTROL;

	config.server_user =		ENV_DEF_USER;

	config.buf_size_header =	ENV_DEF_BUF_HEADER;
	config.buf_size_recv =		ENV_DEF_BUF_RECV;
	config.buf_size_send =		ENV_DEF_BUF_SEND;

	config.db_path =		ENV_DEF_DB_PATH;
	config.rep_path =		ENV_DEF_DB_REP_PATH;		//

	config.http_use =		ENV_DEF_HTTP_USE;
	config.http_root =		ENV_DEF_HTTP_ROOT;
	config.http_timeout =		ENV_DEF_HTTP_TIMEOUT;

	config.webdav =			ENV_DEF_WEBDAV;

	config.log_path =		ENV_DEF_LOG_PATH;		//
	
	return;
}

int parse_config( int argc, char **argv )
{
	extern struct config_t		config;
	
	int				c;		//getopt
	extern char			*optarg;	//getopt
	extern int			optind;		//getopt
	
	regex_t				reg_node;
	regmatch_t			node[5];
	size_t				node_cnt = 5;
	int				match_res;
	int				found = ENV_MAX;
	
	FILE				*conf;
	char				line[1024];

	struct config_node_t		set[ENV_MAX];
	int				set_len = 0;

	optind = 1;

	while ( (c = getopt( argc, argv, "c:hv" )) > 0 )
	{
		switch ( c )
		{
			case 'c':
				config.filepath = malloc( 1024 );
				config.filepath = strncpy( config.filepath, optarg, 1024 );
			break;
			case 'h':
				printf( "Usage: %s [-c configfile] [-p port] [-d datadir]\n", basename(argv[0]) );
				return -1;
			break;
			case 'v':
				printf( "%s\n", SERVER_SOFTWARE );
				return -1;
			break;
		}
	}

	conf = fopen( config.filepath, "r" );
	if ( conf == NULL )
	{
		chdir( __PREFIX );
		conf = fopen( config.filepath, "r" );
		if ( conf == NULL )
		{
			perror("Cannot open config file");
			return -1;
		}
	}

	regcomp( &reg_node,	"\\s*([^=]*)\\s*=\\s?((\"[^\"]*\")|(.*))", REG_EXTENDED | REG_ICASE );
	while ( feof( conf ) == 0 && found > 0 )
	{
		fgets( line, 1024, conf );
		match_res = regexec( &reg_node, line, node_cnt, node, 0 );

		if ( match_res == 0 )
		{
			if ( line[node[0].rm_so] == '#' )
				continue;

			//printf("(%d, %d, %d, %d, %d, %d, %d, %d, %d, %d)%s", node[0].rm_so, node[0].rm_eo, node[1].rm_so, node[1].rm_eo, node[2].rm_so, node[2].rm_eo, node[3].rm_so, node[3].rm_eo, node[4].rm_so, node[4].rm_eo, line);

			set[set_len].buf = malloc( node[0].rm_eo+1 );
			snprintf( set[set_len].buf, node[0].rm_eo, "%s", line );

			set[set_len].field = &(set[set_len].buf[node[1].rm_so]);
			set[set_len].buf[node[1].rm_eo] = '\0';
			set[set_len].field_len = node[1].rm_eo-node[1].rm_so;

			set[set_len].value = &(set[set_len].buf[node[2].rm_so]);
			set[set_len].buf[node[2].rm_eo] = '\0';
			set[set_len].value_len = node[2].rm_eo-node[2].rm_so;

			if ( set[set_len].value[0] == '"' && set[set_len].value[set[set_len].value_len-2] == '"' )
			{
				set[set_len].value_len--;
				set[set_len].value[set[set_len].value_len-1] = '\0';
				set[set_len].value++;
			}

			set_len++;
			found--;
		}
	}
	regfree( &reg_node );

	fclose( conf );

	config.server_port = get_config_int( set, set_len, "port", config.server_port );
	config.max_session = get_config_int( set, set_len, "max session", config.max_session );
	config.poll_timer = get_config_int( set, set_len, "poll timer", config.poll_timer );
	config.buf_size_header = get_config_int( set, set_len, "header buf size", config.buf_size_header );
	config.buf_size_recv = get_config_int( set, set_len, "recv buf size", config.buf_size_recv );
	config.buf_size_send = get_config_int( set, set_len, "send buf size", config.buf_size_send );
	config.http_timeout = get_config_int( set, set_len, "timeout", config.http_timeout );
	config.access_control = get_config_bool( set, set_len, "access control", config.access_control );

	config.server_user = get_config_char( set, set_len, "server user", config.server_user );
	config.db_path = get_config_char( set, set_len, "data dir", config.db_path );
	config.rep_path = get_config_char( set, set_len, "repl dir", config.rep_path );
	config.http_root = get_config_char( set, set_len, "http document root", config.http_root );
	config.log_path = get_config_char( set, set_len, "log dir", config.log_path );

	config.http_use = get_config_bool( set, set_len, "http use", config.http_use );
	config.webdav = get_config_bool( set, set_len, "webdav use", config.webdav );

	while ( (c = getopt( argc, argv, "p:d:" )) > 0 )
	{
		switch ( c )
		{
			case 'p':
				config.server_port = atoi(optarg);
			break;
			case 'd':
				config.db_path = malloc( 1024 );
				config.db_path = strncpy( config.db_path, optarg, 1024 );
			break;
		}
	}

	for ( ; set_len--; )
		free( set[set_len].buf );

	return 0;
}

char *get_config_char( struct config_node_t *set, int set_len, char *find, char *def )
{
	int		i;
	int		len;

	len = strlen( find );

	for ( i=set_len; i--; )
	{
		if ( strncasecmp( set[i].field, find, len ) == 0 )
			return set[i].value;
	}

	return def;
}

int get_config_int( struct config_node_t *set, int set_len, char *find, int def )
{
	int		i;
	int		len;

	len = strlen( find );

	for ( i=set_len; i--; )
	{
		if ( strncasecmp( set[i].field, find, len ) == 0 )
			return atoi(set[i].value);
	}

	return def;
}

short get_config_bool( struct config_node_t *set, int set_len, char *find, short def )
{
	int		i;
	int		len;

	len = strlen( find );

	for ( i=set_len; i--; )
	{
		if ( strncasecmp( set[i].field, find, len ) == 0 )
		{
			if ( strncmp( set[i].value, "1", 1 ) == 0 || strncasecmp( set[i].value, "true", 4 ) == 0 )
				return TRUE;
			else
				return FALSE;
		}
	}

	return def;
}

