#include "include/common.h"

void monitor_thread_push( struct thread_t *self )
{
	extern struct stat_t	thread_stat;
	struct thread_t		*next;

	self->next = NULL;

	pthread_mutex_lock( &thread_stat.lock );

	if ( thread_stat.first == NULL )
		thread_stat.first = self;
	else
	{
		for ( next = thread_stat.first; next != NULL; next=next->next )
		{
			if ( next->next == NULL )
			{
				next->next = self;
				break;
			}
		}
	}
	thread_stat.count++;
	pthread_mutex_unlock( &thread_stat.lock );
}

void monitor_thread_delete( struct thread_t *self )
{
	extern struct stat_t	thread_stat;
	struct thread_t		*next;

	pthread_mutex_lock( &thread_stat.lock );

	if ( thread_stat.first == self )
	{
		if ( self->next != NULL )
			thread_stat.first = self->next;
		else
			thread_stat.first = NULL;
	}
	else
	{
		for ( next = thread_stat.first; next != NULL; next=next->next )
		{
			if ( next->next == self )
			{
				if ( self->next )
					next->next = self->next;
				else
					next->next = NULL;
				break;
			}
		}
	}
	thread_stat.count--;
	pthread_mutex_unlock( &thread_stat.lock );
}

#define R_BUF		&(self->response.buf[self->response.offset])
#define R_BUF_LIMIT	(self->response.buf_size-self->response.offset)
int monitor_response( struct thread_t *self )
{
	extern struct stat_t	thread_stat;

	long long int		rowid;
	struct thread_t		*next;
	time_t			now;

	now = time( &now );

	rowid = db_last_rowid( &self->db );

	pthread_mutex_lock( &thread_stat.lock );

	self->response.start = self->response.offset = 20;
	self->response.offset += response_buf_open( R_BUF, R_BUF_LIMIT, 0, rowid, thread_stat.count, &self->bench_start );

	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_OPEN );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "process" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "db" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "user" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "status" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "sql" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "since_init" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, "since_last" );
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_CLOSE );

	for ( next = thread_stat.first; next != NULL; next=next->next )
	{
		//"<session><process>%d</process><db>%s</db><user>%s</user><status>%s</status><worksql>%s</worksql><since><init>%d</init><idle>%d</idle></since></session>\n"
		self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_MONITOR, next->tid, next->db.db_name, next->user.id,
			(next->status == THREAD_INIT?"Initialize":(next->status == THREAD_WAIT?"Client Wainting":"DBMS Working")),
			next->request.data, now-next->time_init, now-next->time_last );
	}
	self->response.offset += sprintf( &(self->response.buf[self->response.offset]), XML_RESPONSE_CLOSE );
	pthread_mutex_unlock( &thread_stat.lock );

	self->response.status = 200;
	self->response.status_msg = "OK";

	return response( self );
}
#undef R_BUF
#undef R_BUF_LIMIT
