#include "include/common.h"

#define	CLIENT_KEEP_OR_BREAK	{if ( self.keep_alive == TRUE ) {	\
					client_session_init( &self );	\
					goto request_continue;		\
				}else break;}

void *client( void *data )
{
	extern struct config_t	config;

	int			client_fd = *((int *) data);
	struct thread_t		self;

	struct pollfd		pollfd;
	int			poll_res;
	int			read_res;

	int			auth_res;

	int			db_res;
	int			result;

	int			http_res;


	memset( &self, 0, sizeof(self) );
	self.init = FALSE;
	self.status = THREAD_INIT;

	time( &self.time_init );
	self.time_last = 0;

	if ( client_fd <= 0 )
		pthread_exit(0);

	self.client_fd = client_fd;

	self.request.buf_size = config.buf_size_recv;
	self.request.buf = malloc( self.request.buf_size );

	self.response.header_buf = malloc( config.buf_size_header );

	self.response.buf_size = config.buf_size_send;
	self.response.buf = malloc( self.response.buf_size );

	self.tid = pthread_self();

	pollfd.fd = client_fd;
	pollfd.events = POLLIN | POLLPRI | POLLERR | POLLHUP;

	monitor_thread_push( &self );
	self.init = TRUE;

	while ( 1 )
	{
request_continue:
		self.status = THREAD_WAIT;
		poll_res = poll( &pollfd, 1, config.http_timeout*1000 );

		if ( poll_res <= 0 )
			break;

		if ( self.request.buf_size-self.request.read_len <= 1024*8 )
		{
			self.request.buf = realloc( self.request.buf, self.request.buf_size + LIMIT_BUF_INCREMENT );
			self.request.buf_size += LIMIT_BUF_INCREMENT;
		}

		read_res = read( self.client_fd, &(self.request.buf[self.request.read_len]), self.request.buf_size-self.request.read_len );

		if ( read_res <= 0 )
			break;

		self.request.read_len += read_res;

		if ( self.request.header_len <= 0 )
		{
			self.request.buf[self.request.read_len] = '\0';
#ifdef __DEBUG
			printf("%s\n", self.request.buf);
#endif
			if ( request_header_parse( &self.request ) == SQLITE_ERROR )
			{
				response_error( &self, RESPONSE_ERR_BAD_REQUEST, NULL );
				goto client_error;
			}
		}

		if ( self.request.header_len > 0 ) 
		{
			switch ( request_end( &self.request ) )
			{
				case SQLITE_OK:
				break;
				case SQLITE_EMPTY:
					goto request_continue;
				break;
				case SQLITE_ERROR:
					if ( self.request.header.method == REQUEST_METHOD_GET )
						break;
					response_error( &self, RESPONSE_ERR_BAD_REQUEST, NULL );
					goto client_error;
				break;
			}

			self.request.buf[self.request.read_len] = '\0';

			self.keep_alive = request_keep_alive( &self.request );

			self.status = THREAD_WORK;

			if ( self.request.header.method == REQUEST_METHOD_POST || self.request.header.method == REQUEST_METHOD_OPTION )
			{
				time( &self.time_last );
				gettimeofday( &self.bench_start, NULL );

				auth_res = client_auth_proc( &self );

				if ( auth_res != SQLITE_OK )
				{
					//auth error
					if ( response_error( &self, RESPONSE_ERR_UNAUTHORIZED, NULL ) != SQLITE_OK )
						break;
					CLIENT_KEEP_OR_BREAK
				}

				db_res = client_open_db( &self );

				if ( self.request.header.method == REQUEST_METHOD_POST )
				{
					if ( db_res == SQLITE_NOTFOUND || db_res == SQLITE_OK )
					{
						result = extend_process( &self, db_res );
						if ( result == SQLITE_OK )
							CLIENT_KEEP_OR_BREAK
						else if ( result == SQLITE_ERROR_SEND )
							break;

						if ( self.user.super == 1 )
						{
							result = admin_process( &self, db_res );
							if ( result == SQLITE_OK )
								CLIENT_KEEP_OR_BREAK
							else if ( result == SQLITE_ERROR_SEND )
								break;
						}
					}
				}

				if ( db_res == SQLITE_OK )
				{
					//db process
					if ( self.request.data_len <= 0 )
					{
						if ( response_info( &self, RESPONSE_INFO_LOGIN_SUCCESS, NULL ) != SQLITE_OK )
							break;
						CLIENT_KEEP_OR_BREAK
					}

					result = client_process( &self, self.request.data );
					if ( result == SQLITE_OK )
						CLIENT_KEEP_OR_BREAK
					else if ( result == SQLITE_ERROR_SEND )
						break;
				}
				else
				{
					if ( response_error( &self, RESPONSE_ERR_DB_FOUND, NULL ) != SQLITE_OK )
						break;
					CLIENT_KEEP_OR_BREAK
				}
			}
			else if ( self.request.header.method == REQUEST_METHOD_PUT || self.request.header.method == REQUEST_METHOD_DELETE )
			{
				auth_res = client_auth_proc( &self );

				if ( auth_res != SQLITE_OK || self.user.super != 1 )
				{
					//auth error
					if ( response_error( &self, RESPONSE_ERR_UNAUTHORIZED, NULL ) != SQLITE_OK )
						break;
					CLIENT_KEEP_OR_BREAK
				}

				if ( config.webdav == TRUE )
				{
					http_res = webdav_proc( &self );
					if ( http_res == SQLITE_OK )
					{
						CLIENT_KEEP_OR_BREAK
					}
					else
						break;
				}
				else
				{
					if ( response_error( &self, RESPONSE_ERR_BAD_REQUEST, NULL ) != SQLITE_OK )
						break;
					CLIENT_KEEP_OR_BREAK
				}
			}
			else
			{
				if ( config.http_use == TRUE )
				{
					http_res = http_proc( &self );
					if ( http_res == SQLITE_OK )
					{
						CLIENT_KEEP_OR_BREAK
					}
					else
						break;
				}
				else
				{
					if ( response_error( &self, RESPONSE_ERR_BAD_REQUEST, NULL ) != SQLITE_OK )
						break;
					CLIENT_KEEP_OR_BREAK
				}
			}
			CLIENT_KEEP_OR_BREAK
		}
	}

client_error:

#ifdef __DEBUG
	printf("***** Client thread terminated.\n");
#endif
	client_safe_exit( &self );
	
	pthread_exit( 0 );
}

int client_auth_proc( struct thread_t *self )
{
	char				*cookie_value;
	short				cookie_auth;
	char				*base64;

	char				iddata[1024];
	int				iddata_len;

	char				*id;
	char				*pass;

	struct timeval			now;
	char				tmp[64];
	int				tmp_len;

	cookie_auth = FALSE;

	if ( self->request.header.cookie != NULL && strlen(self->session_key) > 0 )
	{
		cookie_value = strstr( self->request.header.cookie, "SQLSESSION=" );
		if ( cookie_value != NULL )
		{
			cookie_value += 11;
			if ( strncmp( cookie_value, self->session_key, MINVAL( 32, MINVAL( strlen(cookie_value), strlen(self->session_key) ) ) ) == 0 )
			{
				return SQLITE_OK;
			}
		}
	}

	if ( self->request.header.auth == NULL )
		return SQLITE_AUTH;

	base64 = strchr( self->request.header.auth, ' ' );
	if ( base64 == NULL )
		return SQLITE_AUTH;

	iddata_len = base64_decode( iddata, ++base64, strlen(base64) );
	if ( iddata_len <= 0 )
		return SQLITE_AUTH;
	iddata[iddata_len] = '\0';

	id = iddata;
	pass = strchr( iddata, ':' );
	if ( pass == NULL )
		return SQLITE_AUTH;
	pass[0] = '\0';
	pass++;

	if ( auth_user_check( id, pass, &self->user ) == SQLITE_OK )
	{
		gettimeofday( &now, NULL );
		tmp_len = snprintf( tmp, 64, "%ld%ld", now.tv_usec, now.tv_sec );
		for ( ; tmp_len--; )
			tmp[tmp_len] += ( 17 + tmp[tmp_len]%(tmp_len+1) );

		strncpy( self->session_key, tmp, 32 );
		self->response.set_cookie = 1;
		return SQLITE_OK;
	}
	else
		return SQLITE_AUTH;
}

int client_open_db( struct thread_t *self )
{
	extern struct config_t	config;
	int			open_res;
	
	if ( strlen( self->request.header.db_name ) <= 0 )
		return SQLITE_NOTFOUND;

	self->request.header.extend = strchr( self->request.header.db_name, '/' );
	if ( self->request.header.extend != NULL )
	{
		self->request.header.extend[0] = '\0';
		self->request.header.extend++;
		self->request.header.db_name_len = strlen( self->request.header.db_name );
	}

	if ( strlen(self->db.db_name) == self->request.header.db_name_len
		&& strncasecmp( self->db.db_name, self->request.header.db_name, self->request.header.db_name_len ) == 0 )
	{
		return SQLITE_OK;
	}

	db_close( &self->db );
	open_res = db_open( &self->db, self->request.header.db_name );

	if ( open_res == SQLITE_OK )
	{
		memcpy( self->db.db_name, self->request.header.db_name, self->request.header.db_name_len );
		if ( config.access_control == TRUE )
		{
			self->user.dbname = self->db.db_name;
			db_set_authorizer( &self->db, &(self->user) );
		}
		return SQLITE_OK;
	}
	return open_res;
}

#define R_BUF		&(self->response.buf[self->response.offset])
#define R_BUF_LIMIT	(self->response.buf_size-self->response.offset)
int client_process( struct thread_t *self, char *query )
{
	int		result;
	int		change;
	long long int	rowid;
	int		rows;
	int		columns;
	char		*column_name;
	int		column_name_len;
	int		i, p;
	char		*data;
	int		data_len;

#ifdef __DEBUG
	printf("***** SQL: %s\n", query);
#endif
	result = db_get_table( &self->db, query );

	if ( result != SQLITE_OK )
		return response_error( self, RESPONSE_ERR_INTERNAL, self->db.errmsg );

	change = db_changes( &self->db );
	rowid = db_last_rowid( &self->db );
	rows = db_table_num_rows( &self->db );

	self->response.start = self->response.offset = 20;

	self->response.offset += response_buf_open( R_BUF, R_BUF_LIMIT, change, rowid, rows, &self->bench_start );

	self->response.status = 200;
	self->response.status_msg = "OK";
	//query success
	if ( db_table_has_rows( &self->db ) == SQLITE_OK )
	{
		columns = db_table_columns( &self->db );
		//columns open
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_OPEN );
		for ( i=0; i < columns; i++ )
		{
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
			column_name_len = db_table_column_name( &column_name, &self->db, i );
			self->response.offset += response_xml_printf( R_BUF, column_name, column_name_len );
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
		}
		self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_CLOSE );

		for ( i=0; i < rows; i++ )
		{
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_OPEN );
			for ( p=0; p < columns; p++ )
			{
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_OPEN );
				data = db_table_result( &self->db, i, p );
				if ( data != NULL )
				{
					data_len = strlen( data );
					if ( data_len + 100 > R_BUF_LIMIT )
					{
						/* chunked response start */
						result = response_flush( self, FALSE );
						if ( result != SQLITE_OK )
						{
							db_free_table( &self->db );
							return result;
						}
					}
					self->response.offset += response_xml_printf( R_BUF, data, data_len );
				}
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_DATA_CLOSE );
			}
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_ROW_CLOSE );
		}
	}
	else
	{
		columns = db_table_columns( &self->db );
		if ( columns > 0 )
		{
			//columns open
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_OPEN );
			for ( i=0; i < columns; i++ )
			{
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_OPEN );
				column_name_len = db_table_column_name( &column_name, &self->db, i );
				self->response.offset += response_xml_printf( R_BUF, column_name, column_name_len );
				self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMN_CLOSE );
			}
			self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_COLUMNS_CLOSE );
		}
	}

	self->response.offset += snprintf( R_BUF, R_BUF_LIMIT, XML_RESPONSE_CLOSE );
	db_free_table( &self->db );

	if ( self->response.chunked == TRUE )
		return response_flush( self, TRUE );
	else
		return response( self );
}
#undef R_BUF
#undef R_BUF_LIMIT

void client_session_init( struct thread_t *self )
{
	self->keep_alive = FALSE;

	self->request.read_len = self->request.header_len = 0;
	self->request.data_len = /*self->request.header.node_idx = */self->request.header.method = self->request.header.db_name_len = 0;
	self->request.data = NULL;
	self->request.header.db_name = self->request.header.extend = NULL;

	self->request.header.cookie = self->request.header.auth = self->request.header.modified = self->request.header.connection = NULL;
	self->request.header.content_length = -1;

	self->response.start = self->response.offset = 0;
	self->response.status = 0;
	self->response.header_sent = self->response.set_cookie = self->response.chunked = FALSE;
	self->response.status_msg = self->response.errmsg = NULL;
}

void client_safe_exit( struct thread_t *self )
{
	if ( self->request.buf != NULL )
		free( self->request.buf );
	if ( self->response.header_buf != NULL )
		free( self->response.header_buf );
	if ( self->response.buf != NULL )
		free( self->response.buf );
	close( self->client_fd );

	if ( self->db.db != NULL )
	{
		if ( self->db.auth == TRUE && self->db.transaction == TRUE )
		{
			if ( self->db.auto_commit == TRUE )
				db_commit( &self->db );
			else
				db_rollback( &self->db );
		}
		db_close( &self->db );
	}

	if ( self->init == TRUE )
		monitor_thread_delete( self );
}

