/* DB auth */
struct db_user_t {
	char			id[128];
	char			pass[128];
	short			super;
	char			*dbname;
};

struct _auth_data_t {
	char			**table;
	int			rows;
	int			cols;
};

struct auth_t {
	struct _auth_data_t	user;
	struct _auth_data_t	db;
	struct _auth_data_t	object;
};

struct grant_t {
	char			*rule;
	char			*dbname;
	char			*object;
	char			*user;
};



typedef union {
	struct sockaddr		sa;
	struct sockaddr_in	sa_in;
} sock_t;

struct config_node_t {
	char			*buf;
	char			*field;
	int			field_len;
	char			*value;
	int			value_len;
};

struct config_t {
	char			*filepath;
	int			server_port;
	int			max_session;

	int			poll_timer;
	short			access_control;

	char			*server_user;

	int			buf_size_header;
	int			buf_size_recv;
	int			buf_size_send;

	char			*db_path;
	char			*rep_path;

	short			http_use;
	char			*http_root;
	int			http_timeout;

	short			webdav;

	char			*log_path;
};

struct request_header_t {
	int			method;
	char			*db_name;
	char			*extend;
	int			db_name_len;

	char			*query;

	char			*cookie;
	char			*auth;
	char			*modified;
	char			*connection;
	int			content_length;
};

struct request_t {
	char			*buf;
	int			buf_size;
	int			read_len;

	struct request_header_t	header;
	int			header_len;

	char			*data;
	int			data_len;
};

struct response_t {
	char			*header_buf;

	char			*buf;
	int			buf_size;
	int			offset;
	int			start;

	short			header_sent;
	short			set_cookie;
	short			chunked;

	short			status;
	char			*status_msg;
	char			*errmsg;

	struct stat		stat;
	int			file;
	off_t			file_offset;
};

struct pstmt_t {
	char			*sql;
	char			name[128];
	sqlite3_stmt		*stmt;
};

struct db_t {
	sqlite3			*db;					//sqlite3 db pointer
	char			db_name[LIMIT_MAX_DBNAME];		//sqlite3 db name, db filename exclude ".db"
	char			db_filename[LIMIT_MAX_DBFILENAME];	//sqlite3 db full file name
	short			opened;
	char			*errmsg;				//sqlite3 lib error msg

	sqlite3_stmt		*stmt;
	const char		**sql_tail;

	struct pstmt_t		*pstmt;
	int			pstmt_cnt;

	char			**table;
	int			table_res;
	int			table_nrow;
	int			table_ncol;

	short			transaction;
	short			auto_commit;

	short			auth;
};

struct thread_t {
	/*
	pthread_t		thread;
	pthread_mutex_t		thread_mutex;
	pthread_cond_t		thread_cond;
	*/
	short			init;
	short			status;
	time_t			time_init;
	time_t			time_last;

	int			client_fd;

	struct request_t	request;
	struct response_t	response;

	struct db_user_t	user;
	char			session_key[32];

	short			keep_alive;

	struct db_t		db;

	/** thread stat */
	pthread_t		tid;
	struct thread_t		*next;

	/** bench */
	struct timeval		bench_start;
};

struct stat_t {
	struct thread_t		*first;
	pthread_mutex_t		lock;
	int			count;
};

struct desc_t {
	char const		*type;
	char const		*seq;
	int			notnull;
	int			pkey;
	int			autoincrement;
};

