#include "sqlited.h"
#include "sqlited_internal.h"

sqlited_result *sqlited_query( sqlited *db, const char *query )
{
	extern int	errno;

	size_t		query_len;
	int		result;

	query_len = strlen( query );

	if ( db->send_buf == NULL || db->send_buf_size <= 0 )
	{
		db->send_buf_size = MAXVAL( SQLITED_DEF_SEND_BUF, (query_len+1) * 3 );
		db->send_buf = (char *) malloc( db->send_buf_size );
	}

	if ( db->send_buf_size <= query_len * 3 )
	{
		db->send_buf_size = (query_len+1) * 3;
		db->send_buf = (char *) realloc( db->send_buf, db->send_buf_size );
	}

	db->send_length = _sqlited_make_body( db, query, query_len );
	db->send_length += _sqlited_make_header( db, SQLITED_METHOD_POST );

	result = _sqlited_request( db, &(db->send_buf[db->send_offset]), db->send_length );
	if ( result != SQLITED_OK )
	{
		db->errmsg = strerror( errno );
		return NULL;
	}

	result = _sqlited_response( db );
	if ( result == SQLITED_RESPONSE_END )
		return _sqlited_response_parse( db );

	//printf("%s\n", db->recv_buf);

	db->errmsg = strerror( errno );
	return NULL;
}

int _sqlited_request( sqlited *db, const char *buf, sqlited_int64 length )
{
	sqlited_int64	offset = 0;
	int		res;
	struct pollfd	pollfd;

	pollfd.fd = db->sock;
	pollfd.events = POLLOUT;
	while ( offset < length )
	{
		res = poll( &pollfd, 1, 1000 );
		if ( res > 0 )
		{
			res = write( db->sock, &(buf[offset]), length-offset );
			if ( res < 0 )
				return SQLITED_ERROR_SEND;
			offset += res;
		}
		else if ( res < -1 )
			return SQLITED_ERROR_SEND;
	}

	return SQLITED_OK;
}

const unsigned char url_encode_hex[16] = { "0123456789ABCDEF" };
sqlited_int64 _sqlited_make_body( sqlited *db, const char *query, size_t query_len )
{
	size_t	length;
	int	i;
	int	c;
	char	*tmp;

	length = query_len;

	if ( query == '\0' )
		return db->send_length;

	tmp = (char *) query;
	i = db->send_length = 1024;
	c = 0;

	while ( c < query_len )
	{
		if ( isalnum( tmp[0] ) || tmp[0] == '-' || tmp[0] == '=' )
			db->send_buf[i++] = tmp[0];
		else if ( tmp[0] == ' ' )
			db->send_buf[i++] = '+';
		else
		{
			db->send_buf[i++] = '%';
			db->send_buf[i++] = url_encode_hex[ tmp[0]/16 ];
			db->send_buf[i++] = url_encode_hex[ tmp[0]%16 ];
		}
		c++;
		tmp++;
	}

	db->send_buf[i] = '\0';

	return i;
}

sqlited_int64 _sqlited_make_header( sqlited *db, short method )
{
	size_t	length;
	if ( method == SQLITED_METHOD_POST )
	{
		if ( db->session_key != NULL )
			length = snprintf( db->send_header, 1024, "POST /%s HTTP/1.1\r\nCookie: SQLSESSION=%s\r\nContent-length: %lld\r\nConnection: Keep-Alive\r\n\r\n", db->dbname, db->session_key, db->send_length );
		else
			length = snprintf( db->send_header, 1024, "POST /%s HTTP/1.1\r\nAuthorization: Basic %s\r\nContent-length: %lld\r\nConnection: Keep-Alive\r\n\r\n", db->dbname, db->auth_str, db->send_length );
		db->send_offset = 1024-length;
		db->send_length += length;
		memcpy( &(db->send_buf[db->send_offset]), db->send_header, length );
		return length;
	}

	return -1;
}

void _sqlited_request_session_init( sqlited *db )
{
	db->send_length = db->recv_length = 0;
	db->response.header_end = db->response.end = 0;
	db->response.chunk_offset = NULL;
}
